/*
btsplit.c Version 0.1.0  Btree split non-root node
Copyright (C) 2003  dondalah@ripco.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:

	Free Software Foundation, Inc.
	59 Temple Place - Suite 330
	Boston, MA  02111-1307, USA.
*/

/* See spltroot.c for splitting root node */

#include <bthash.h>

void btsplit(bthdr)
btreefmt *bthdr;
   {
   int currcd;
   int rslt;
   bt_nodefmt *node;
   bt_nodefmt *midnode;
   unsigned char *p,*q,*r;
   unsigned char *rcdptr;
   bthdr->reason = 0;
   node = (bt_nodefmt *) p = (unsigned char *)
      bthdr->blk + bthdr->midofst;

   midnode = (bt_nodefmt *) malloc(bthdr->rcdnode+16);
   if (midnode == NULL)
      {
      fprintf(stderr,"btsplit: out of "
	 "memory allocating midnode\n");
      bthdr->reason = BTREE_ERR + 1701;
      return;
      } /* out of mem */
   p = (unsigned char *) bthdr->blk + bthdr->midofst;
   q = (unsigned char *) p + bthdr->rcdnode;
   r = (unsigned char *) midnode;
   while (p < q) *r++ = *p++;

   bthdr->ofstleft = bthdr->tail->prev->ofst;
   p = (unsigned char *) bthdr->blk + bthdr->blkhdr; 
   q = (unsigned char *) bthdr->blk + bthdr->midofst;
   r = (unsigned char *) bthdr->blkleft + bthdr->blkhdr; 
   while (p < q) *r++ = *p++;
   p = (unsigned char *) r;
   q = (unsigned char *) bthdr->blkleft + bthdr->blksz;
   while (p < q) *p++ = '\0';
   bthdr->blkleft->count = bthdr->nodesz >> 1;
   bthdr->blkleft->right = (off_t) midnode->left;
   btputblk(bthdr,bthdr->ofstleft,bthdr->blkleft);
   if (bthdr->reason)
      {
      free(midnode);
      return;
      } /* if btputblk err */
   p = (unsigned char *) bthdr->blkleft; 
   q = (unsigned char *) p + bthdr->blksz;
   while (p < q) *p++ = '\0';

   bthdr->ofstrght = btnewblk(bthdr);
   if (bthdr->reason)
      {
      free(midnode);
      return;
      } /* if btnewblk err */
   p = (unsigned char *) bthdr->blk + bthdr->midofst
      + bthdr->rcdnode; 
   q = (unsigned char *) bthdr->blk + bthdr->blksz;
   r = (unsigned char *) bthdr->blkrght + bthdr->blkhdr; 
   while (p < q) *r++ = *p++;
   p = (unsigned char *) r;
   q = (unsigned char *) bthdr->blkrght + bthdr->blksz;
   while (p < q) *p++ = '\0';
   bthdr->blkrght->right = (off_t) bthdr->blk->right;
   bthdr->blkrght->count = bthdr->nodesz >> 1;
   btputblk(bthdr,bthdr->ofstrght,bthdr->blkrght);
   if (bthdr->reason)
      {
      free(midnode);
      return;
      } /* if btputblk err */
   p = (unsigned char *) bthdr->blkrght; 
   q = (unsigned char *) p + bthdr->blksz;
   while (p < q) *p++ = '\0';

   btpop(bthdr);
   if (bthdr->reason)
      {
      free(midnode);
      return;
      } /* if btpop err */
   btgetblk(bthdr,bthdr->tail->prev->ofst);
   if (bthdr->reason)
      {
      free(midnode);
      return;
      } /* if btgetblk err */
   currcd = 1;
   rcdptr = (unsigned char *) bthdr->blk + bthdr->blkhdr;
   while (currcd <= bthdr->blk->count)
      {
      p = (unsigned char *) rcdptr
	 + bthdr->rcdhdr
	 + bthdr->keyofst;
      q = (unsigned char *) p + bthdr->keylen;
      r = (unsigned char *) midnode
	 + bthdr->rcdhdr + bthdr->keyofst;
      while (p < q)
	 {
	 rslt = *r - *p;
	 if (rslt) break;
	 p++;
	 r++;
	 } /* for each byte in key */
      if (!rslt)
	 {
	 unsigned char *p;
	 fprintf(stderr,"btsplit: duplicate key\n");
	 p = (unsigned char *) midnode
	    + bthdr->rcdhdr
	    + bthdr->keyofst
	    + bthdr->keylen;
	 *p = '\0';
	 fprintf(stderr,"%s\n", p);
         bthdr->reason = BTREE_ERR + 1702;
         free(midnode);
         return;
	 } /* dup rcd key */
      else if (rslt < 0) break;
      currcd++;
      rcdptr = (unsigned char *) rcdptr
	 + bthdr->rcdhdr
	 + bthdr->rcdlen;
      } /* for each record in block */

   if (rslt < 0)
      {
      p = (unsigned char *) rcdptr;
      q = (unsigned char *) p
         + ((bthdr->blk->count - currcd + 1)
	    * bthdr->rcdnode);
      r = (unsigned char *) bthdr->blkleft;
      while (p < q) *r++ = *p++;

      midnode->left = bthdr->ofstleft;
      p = (unsigned char *) midnode; 
      q = (unsigned char *) p + bthdr->rcdnode;
      r = (unsigned char *) rcdptr;
      while (p < q) *r++ = *p++;

      node = (bt_nodefmt *) bthdr->blkleft;
      node->left = bthdr->ofstrght;
      p = (unsigned char *) r;
      q = (unsigned char *) p
         + ((bthdr->blk->count - currcd + 1)
	    * bthdr->rcdnode);
      r = (unsigned char *) bthdr->blkleft;
      while (p < q) *p++ = *r++;

      p = (unsigned char *) bthdr->blkleft;
      q = (unsigned char *) p
         + ((bthdr->blk->count - currcd + 1)
	    * bthdr->rcdnode);
      while (p < q) *p++ = '\0';
      } /* if insert before a node (left,left) */
   else if (rslt > 0)
      {
      midnode->left = bthdr->blk->right;
      p = (unsigned char *) midnode; 
      q = (unsigned char *) p + bthdr->rcdnode;
      r = (unsigned char *) rcdptr;
      while (p < q) *r++ = *p++;
      bthdr->blk->right = bthdr->ofstrght;
      } /* if insert at end of block (left,right) */
   else /* rslt == 0 */
      {
      fprintf(stderr,"btsplit: duplicate key "
	 "after compare loop in parent\n");
      fprintf(stderr,"btsplit: ofst %08x ",
	 (unsigned int) bthdr->tail->prev->ofst);
      p = (unsigned char *) midnode + bthdr->rcdhdr;
      *(p + bthdr->rcdlen) = '\0';
      fprintf(stderr,"btsplit: record %s ", p);
      bthdr->reason = BTREE_ERR + 1703;
      free(midnode);
      return;
      } /* if duplicate key */

   bthdr->blk->count++;
   btputblk(bthdr,
      bthdr->tail->prev->ofst,
      bthdr->blk);
   bthdr->ofstleft = (off_t) 0;
   bthdr->ofstrght = (off_t) 0;
   free(midnode);
   } /* btsplit */
